package sdlip.examples;

import sdlip.SDLIPException;
/* holder for integer values */
import org.omg.CORBA.IntHolder;
/* transport module libraries */
import sdlip.helpers.*;
/* following packages are imported only if DOM-aware XMLObjects are used */
import sdlip.xml.dom.*;
import org.w3c.dom.*;
/* SDLIP XML element names */
import sdlip.SDLIP;

/**
 * This is the SimpleClient application described in the SDLIP Java tutorial.
 */


public class SimpleClient {

  public static void main(String args[]) {

    if(args.length != 2) {
      System.err.println("Usage: java sdlip.examples.SimpleClient <URL_or_IOR> <query text>");
      System.exit(1);
    }

    try {
      /* extract command line parameters */
      String searchServerURI = args[0];
      String queryStr = args[1];

      /* an SDLIP/CORBA or SDLIP/DASL transport module is created
       * depending on whether a URL or an IOR is supplied */
      ClientTransportModule tm = ClientTransportModule.create(searchServerURI);
      tm.setDBG(new DBG(DBG.SILENT));

      /* the result will be held in a DOM-enabled sdlip.xml.dom.XMLObject */
      XMLObject result = new XMLObject();
      /* total expected number of results for this query */
      IntHolder expectedTotal = new IntHolder();

      tm.search(12345,           /* some client SID */
		null,            /* search the default subcollection */
		new XMLObject("<basicsearch xmlns='DAV:'>" +
			      "  <where><contains>" + queryStr + "</contains></where>" +
			      "</basicsearch>"),
		10,              /* return 10 results only */
		null,            /* retrieve all properties */
		0,               /* do not maintain state */
		null,            /* no query options */
		expectedTotal,   /* place to return expectedTotal */
		new IntHolder(), /* place to return stateTimeout - don't care */
		new IntHolder(), /* place to return serverSID - don't care */
		new XMLObject(), /* place to return serverDelegate - don't care */
		result           /* place to return search result */
		);

      /* get the top XML element from an XMLObject */
      Element searchResult = result.getElement();

      /* iterate over all found documents */
      NodeList docList = DOMUtil.getChildElements(searchResult, SDLIP.doc);
      for(int i=0; i < docList.getLength(); i++) {

	/* print document ID */
        Element eDoc = (Element)docList.item(i);
        System.out.println("DOCUMENT: " + DOMUtil.getChildText(eDoc, SDLIP.DID));
	
	/* print all document properties in format NAME = "VALUE" */
	Element ePropList = DOMUtil.getChild(eDoc, SDLIP.propList);
	NodeList propList = DOMUtil.getChildElements(ePropList, "*");
	for(int j=0; j < propList.getLength(); j++) {
	  Element eProp = (Element)propList.item(j);
	  System.out.println("\t" + eProp.getTagName() + " = \"" + DOMUtil.getText(eProp) + "\"");
	}
      }
      System.out.println("EXPECTED TOTAL: " + expectedTotal.value);

    } catch (SDLIPException sdlip) {
      System.err.println("SDLIP exception (" + sdlip.getCode() + "): " + sdlip.getReason());
    }
  }
}
