package sdlip.examples;

import sdlip.SDLIP;
import sdlip.DAV;

import sdlip.ResultAccess;
import sdlip.Search;
import sdlip.Metadata;
import sdlip.SDLIPException;
/* transport module libraries */
import sdlip.helpers.*;
/* DOM-enabled XMLObject */
import sdlip.xml.dom.*;
import sdlip.DublinCore;
/* holder for integer values */
import org.omg.CORBA.IntHolder;

import org.w3c.dom.*;

import java.util.*;
import java.net.*;

import ORG.oclc.ber.*;
import ORG.oclc.z39.*;
import ORG.oclc.z39.client.*;

/**
<basicsearch xmlns='DAV:' xmlns:dc='http://purl.org/metadata/dublin_core#'>
  <where>
    <eq>
     <prop><dc:Creator/></prop>
     <literal>larson</literal>
    </eq>
  </where>
</basicsearch>
 */

public class Z39LSP implements Search, ResultAccess, Metadata, Timeoutable {

  /** A session corresponds to a standing TCP/IP connection to a Z39.50 server.
   * Only a "default" result set is used within a session */

  class Session {

    int SID;
    Timer timer;
    // need it to be able to set the metadata
    Z39LSP lsp;
    Z39session client;
    private int  gservLength = 0;
    String query;
    boolean searchAttempted;
    boolean interrupted;
    int pendingRequests = 0;
    Hashtable records = new Hashtable(); // Integer -> eDoc element

    public Session(int SID, String query, int timeout, Z39LSP lsp) {

      this.SID = SID;
      this.lsp = lsp;
      this.timer = timeout == 0 ? null : new Timer(timeout, lsp, this);
      this.query = query;
      //      board = new SwitchBoard(query);
    }

    Element getRecord(int i) {

      return (Element)records.get(new Integer(i));
    }

    /** after the invocation, range.get() yields either -1 or leaps to the next non-consec. value */
    int getNextConsecutive(RangeEnumerator range) {
      
      int i = range.get();
      while(range.next() && range.get() == i + 1)
	i++;
      return i;
    }

    public Element getDocs(String docSelection,  sdlip.XMLObject docProps) throws SDLIPException {

      System.err.println("--- getDocs");

      initialize();

      System.err.println("--- init done");

      // get the list of attributes to retrieve
      Vector attrs = NestedAttribute.collectAttributes(XMLObject.create(docProps).getElement());

      Element eSearchResult = XMLObject.getDocument().createElement(SDLIP.SearchResult);

      // we know how many results to expect...
      docSelection = docSelection.trim();
      if(docSelection.endsWith("-"))
	docSelection += getNumDocs();

      RangeEnumerator range = new RangeEnumerator(docSelection);

      range.next();
      int counter = 0;

      System.err.println("--- starting loop");

      while(!interrupted && counter < lsp.MAX_DOCS_PER_FETCH) {

	int first = range.get();
	if(first == -1)
	  break;
	int last = first;
	
	System.err.println("--- getRecords: " + first + "," + last);

	Element eDoc = getRecord(range.get());
	if(eDoc == null) {
	  // record fault, fetch next consecutive records
	  first = range.get();
	  last = getNextConsecutive(range);
	  fetchDocuments(first, last);
	} else
	  range.next();
	
	for(int i = first; i <= last; i++) {
	  eDoc = getRecord(i);
	  if(eDoc != null) {
	    // match against available data
	    eSearchResult.appendChild( NestedAttribute.matchDoc(attrs, eDoc) );
	  }
	  counter++;
	}
      }

      //      board.getDocs(eSearchResult, new RangeEnumerator(docSelection));
      return eSearchResult;
    }

    void fetchDocuments(int first, int last) throws SDLIPException {

      if(last > getNumDocs())
	last = getNumDocs();
      if(first > last)
	return;

      lsp.dbg.println(DBG.SHORT, "Fetching records " + first + " through " + last);

      pendingRequests++;
      BerString req = client.present.Request(0,"default", first, last - first + 1,
					     "F", preferredRecordSyntax, 
					     gservLength, gservLength);
      pendingRequests--;

      if (req == null)
	return;
	    // add header for GServer
      //	    if (gservResource != null)
      //	        System.arraycopy(gservResource, 0, req.record(), 0, 
      //		    gservLength);
      try {
	pendingRequests++;
	client.present.Response(((BerConnect)client.connection).doRequest(req, client.host,client.port));
	pendingRequests--;
      } catch (Exception e) {
	lsp.dbg.println(DBG.SHORT, "Fetching records failed: " + e.getMessage());
	client.reset();
      }

      lsp.dbg.println(DBG.SHORT, "" + client.present.numberOfRecordsReturned + 
		      " records returned");
      switch (client.present.presentStatus)
	    {
		case Z39presentApi.success:
		    break;
		case Z39presentApi.partial_1:
		case Z39presentApi.partial_2:
		case Z39presentApi.partial_3:
/*		case Z39presentApi.partial_4:*/
		  lsp.dbg.println(DBG.SHORT, "Partial results returned");
		  break;
		case Z39presentApi.failure:
		  lsp.dbg.println(DBG.SHORT, "Present failed");
		  break;
	    }

      if (client.present.errorCode != 0)
	{
	  lsp.dbg.println(DBG.SHORT, "Could not retrieve results: Z39.50 errorCode=" + client.present.errorCode +
			  ", message='" + client.present.errorMsg + "'");
	}
      addRecords(first, client.present);
    }

    String toString(DataDir dir) {

      if(dir == null)
	return "";

      StringBuffer str = new StringBuffer();

      DataDir ch = dir.child();
      if(ch != null) {
	while(ch != null) {
	  str.append(toString(ch));
	  ch = ch.next();
	}
      } else { // primitive
	str.append(dir.getString());
      }
      
      return str.toString();
    }

    void appendProp(Element ePropList, String tagName, DataDir dir) {
      if(dir != null)
	ePropList.appendChild( DOMUtil.createTextNode(ePropList.getOwnerDocument(), tagName, toString(dir)) );
    }


    void addRecords(int offset, Z39present present) {

      if (present.numberOfRecordsReturned != 0)
	present.decodeRecords();

      Document doc = XMLObject.getDocument();

      for (int i=0; i<present.numberOfRecordsReturned; i++) {

	int DID = offset + i;
	
	if (present.presentData.elementAt(i) == null)
	  {
	    lsp.dbg.println(DBG.SHORT, "Record " + DID + "absent from returned records!");
	    continue;
	  }
	/*
	  if (output_file != null) {
	  if (present.records[i] instanceof BerString)
	  output_file.write(((BerString)present.records[i]).record());
	  }
	  */
	
	Element eDoc = doc.createElement(SDLIP.doc);
	Element eDID = DOMUtil.createTextNode(doc, SDLIP.DID, String.valueOf(DID));
	eDoc.appendChild(eDID);
	Element ePropList = doc.createElement(SDLIP.propList);
	eDoc.appendChild(ePropList);
	records.put(new Integer(DID), eDoc);
	
	DataDir dir = (DataDir)((DbPresentData)present.presentData.elementAt(i)).data;

	if(dbg.is(DBG.VERBOSE))
	  lsp.dbg.println(DBG.VERBOSE, "DOCUMENT:\n" + dir);

	// 0, 1, 8, 245 - almost the same!

	DataDir sec;

	sec = dir.find(245);
	if(sec != null)
	  appendProp(ePropList, DublinCore.Title, sec.find(1));
	sec = dir.find(100);
	if(sec != null)
	  appendProp(ePropList, DublinCore.Creator, sec.find(1));

	// co-authors
	int occ = 0;
	do {
	  sec = dir.find(700, -1, ++occ);
	  if(sec != null)
	    appendProp(ePropList, DublinCore.Creator, sec.find(1));
	} while (sec != null);

	appendProp(ePropList, DublinCore.Description, dir.find(500));
	appendProp(ePropList, DublinCore.Description, dir.find(504));
	appendProp(ePropList, DublinCore.Description, dir.find(300));

	sec = dir.find(650);
	if(sec != null)
	  appendProp(ePropList, DublinCore.Subject, sec.find(1));
	
	sec = dir.find(260);
	if(sec != null) {
	  appendProp(ePropList, DublinCore.Publisher, sec.find(2));
	  appendProp(ePropList, DublinCore.Date, sec.find(3));
	}
	
	

	//	DataDir dPublisher = dir.find(260);
	/*
	String ids = "";
	DataDir t = dir.child();
	while(t != null) {
	  ids += t.fldid() + ", ";
	  t = t.next();
	}
	  //	DataDir dDescription = dir.find(500);

	//	ePropList.appendChild( DOMUtil.createTextNode(doc, DublinCore.Creator,
	//						      toString(dCreator)) );
	ePropList.appendChild( DOMUtil.createTextNode(doc, DublinCore.Description,
						      ids) );
						      */
	  /*
	  if (present.recordSyntax == null) {
	    System.out.println(((DbPresentData)present.presentData.elementAt(i)).data);
	  }
	  else if (present.recordSyntax.equals(Z39presentApi.SIMPLETEXT_SYNTAX))
	    {
	      DataDir dd = 
		(DataDir)((DbPresentData)present.presentData.elementAt(i)).data;
	      while (dd.form() != ASN1.PRIMITIVE)
		dd = dd.child();
	      if (dd != null)
		System.out.println(dd.dgetChar());
	      
	    }
	  else if (present.recordSyntax.equals(
					       Z39presentApi.OCLC_BER_SYNTAX) || 
		   present.recordSyntax.equals(Z39presentApi.MARC_SYNTAX) ||
		   present.recordSyntax.equals(Z39presentApi.OPAC_SYNTAX))
	    
	    {
	      System.out.println((DataDir)((DbPresentData)present.presentData.elementAt(i)).data);
	    }
	  else if (present.recordSyntax.equals(Z39presentApi.EXPLAIN_SYNTAX))
	    {
	      System.out.println(Explain.getRecord(
						   new DataDir((BerString)((DbPresentData)present.presentData.elementAt(i)).data)));
	    }
	  else
	    System.out.println(((DbPresentData)present.presentData.elementAt(i)).data);
	    */
      }
    }

    public Timer getTimer() {
      return timer;
    }

    public boolean isStateful() {
      return timer != null;
    }

    public int getSID() {
      return SID;
    }

    public int getNumDocs() throws SDLIPException {
      initialize();
      return client.search.resultCount;
    }

    public void cancel() {

      interrupted = true;
      if(timer != null)
	timer.interrupt();

      if(client != null) {
	if(pendingRequests == 0) {
	  // good client
	  BerString req = client.close.Request(0, Z39closeApi.finished, gservLength,
					       gservLength);
	  if (req != null) {
	    // add header for GServer.
	    //	  if (gservResource != null)
	    //		    System.arraycopy(gservResource, 0, req.record(), 0, 
	    //			gservLength);
	    try {client.close.Response(((BerConnect)client.connection).doRequest(req, client.host, client.port)); }
	    catch (Exception e) {}
	    client = null;
	  }
	} else {
	  lsp.dbg.println(DBG.SHORT, "" + pendingRequests + " pending requests, shutting down the connection");
	  client.closeConnection();
	  //	  client = null; otherwise null pointer exception in awaken thread!
	}
      }
    }

    void createSubcollectionInfo() throws SDLIPException {
      
      if(lsp.subcolInfo == null) {

	// determine subcollections
	Document doc = XMLObject.getDocument();
	Element eSubcolInfo = doc.createElement(SDLIP.subcolInfo);
	
	if (client.init.DBList != null &&
	    client.init.DBList.length > 0) {
	  
	  for (int i=0; i<client.init.DBList.length; i++) {
	    
	    Element eSubCol = doc.createElement(SDLIP.subcol);
	    eSubcolInfo.appendChild(eSubCol);
	    eSubCol.appendChild( DOMUtil.createTextNode(doc, SDLIP.subcolName, client.init.DBList[i]) );
	    if (client.init.DisplayDBList != null)
	      eSubCol.appendChild( DOMUtil.createTextNode(doc, SDLIP.subcolName, client.init.DisplayDBList[i]) );
	    if(i == 0) {
	      eSubCol.appendChild( doc.createElement(SDLIP.defaultSubcol) );
	      lsp.defaultSubcol = client.init.DBList[0];
	    }
	  }
	  lsp.subcolInfo = new XMLObject(eSubcolInfo);
	}
      }
    }

    synchronized void initialize() throws SDLIPException {

      if(client == null) {

	lsp.dbg.println(DBG.SHORT, "Initializing Z39.50 session to " + lsp.host + ":" + lsp.port);
	
	client = new Z39session();
	client.host = lsp.host;
	client.port = lsp.port;
	
	//      if(1 == 1)
	//	return;
	
	// init
	pendingRequests++;
	client.initClient(client.host, client.port);
	
	if (client.connection == null) {
	  try {
	    client.connection = new BerConnect(client.host, client.port);
	  } catch (Exception n) {
	    pendingRequests--;
	    client.reset();
	    throw new SDLIPException(SDLIPException.SERVER_ERROR_EXC,
				     "Could not connect to " + client.host + ":" + client.port + ": " + n.getMessage());
	  }
	}
	
	//    System.out.println("Talking to port " + client.port + " of host '" +
	//		       client.host + "'");
	BerString req = client.init.Request(0,null,500000,100000000, client.autho,
					    client.password, "", null, false /*fSetReconnect*/, 
					    gservLength, gservLength);
	pendingRequests--;
	if (req == null)
	  return;
	
	// add header for GServer.
	//	    if (gservResource != null)
	//	        System.arraycopy(gservResource, 0, req.record(), 0, 
	//		    gservLength);
	pendingRequests++;
	try {
	  client.init.Response(((BerConnect)client.connection).doRequest(req, client.host, client.port));
	} catch (Exception e) {
	  pendingRequests--;
	  client.reset();
	  throw new SDLIPException(SDLIPException.SERVER_ERROR_EXC, "Could not initialize " + client.host + ":" + client.port + ": " + e.getMessage());
	}
	pendingRequests--;
	//            System.out.println("Done with Init request");
	//	    if (client.init.MessageOfTheDay != null)
	//		System.out.println(client.init.MessageOfTheDay);
	
	switch (client.init.failureCode) {
	  
	case Z39presentApi.success: // valid init
	  
	  createSubcollectionInfo();
	  break;
	  
	case Z39initApi.InvalidAutho:
	  throw new SDLIPException(SDLIPException.SERVER_ERROR_EXC, "Init failure: Invalid Autho");
	  
	case Z39initApi.BadAuthoPassword:
	  throw new SDLIPException(SDLIPException.SERVER_ERROR_EXC, "Init failure: Unmatched authorization/password");
	  
	case Z39initApi.NoSearchesRemaining:
	  throw new SDLIPException(SDLIPException.SERVER_ERROR_EXC, "Init failure: No Searches remaining for the autho");
	  
	case Z39initApi.IncorrectInterfaceType:
	  throw new SDLIPException(SDLIPException.SERVER_ERROR_EXC, "Init failure: Autho has incorrect interface type");
	  
	case Z39initApi.MaxNumberSimultaneousUsers:
	throw new SDLIPException(SDLIPException.SERVER_ERROR_EXC, "Init failure: Max Number Simultaneous Users for the autho");
	
	case Z39initApi.BlockedIPAddress:
	  throw new SDLIPException(SDLIPException.SERVER_ERROR_EXC, "Init failure: Blocked IP Address");
	  
	default:
	  throw new SDLIPException(SDLIPException.SERVER_ERROR_EXC, "Init failure: Unknown Z39.50 reason code = '" +
				   client.init.failureCode + "'");
	}
	
	/*
	  if (client.init.failureCode != 0)
	  return false;
	  
	  CanDoScan = true;
	  
	  if (client.init.options.length() > 7 && 
	  client.init.options.charAt(7) == 'y') {
	  }
	  else
	  System.out.println(
	"This server indicates that it does NOT support the Scan Facility");
	*/
	lsp.dbg.println(DBG.SHORT, "Session established to " + lsp.host + ":" + lsp.port);
      }

      if(query == null) // collect metadata only
	return;

      if(!searchAttempted) {
	// run the query
	searchAttempted = true;
	lsp.dbg.println(DBG.SHORT, "Running query " + query);

	if(lsp.defaultSubcol == null)
	  throw new SDLIPException(SDLIPException.SERVER_ERROR_EXC, "Internal error: no default subcollection specified!");

	pendingRequests++;
	BerString req = client.search.Request(0, 0,
            	    1, 0,
		    1, "default", lsp.defaultSubcol, 
		    null, null, lsp.preferredRecordSyntax, query, lsp.queryType, 
		    null, null, null, "1.2.840.10003.10.1000.17.7",
		    null, null, gservLength, gservLength);
	pendingRequests--;

	if (req == null)
	  throw new SDLIPException(SDLIPException.SERVER_ERROR_EXC, "Internal error: query translation failed.");

	//		if (gservResource != null)
	//		    System.arraycopy(gservResource, 0, req.record(), 0, 
	//			gservLength);
	pendingRequests++;
	try {
	  client.search.Response(((BerConnect)client.connection).doRequest(req, client.host, client.port));
	} catch (AccessControl aa) {
	  pendingRequests--;
	  throw new SDLIPException(SDLIPException.SERVER_ERROR_EXC, "Access denied.");
	} catch (Exception e) {
	  client.reset();
	  pendingRequests--;
	  throw new SDLIPException(SDLIPException.SERVER_ERROR_EXC, "Search failed: " + e.getMessage());
	}
	pendingRequests--;
	/*
	if (client.search.searchStatus == 0 &&
		    client.search.errorCode == 107 &&
		    query_type == 101)
		{
		    System.out.println("Query Type101 not support by server. "
			+ " Degrading to Type1 and retrying. ");
		    System.out.println("All future queries will be of Type1.");
		    query_type = 1;
		    return true;
		}
		*/
	if (client.search.searchStatus != 0 /*|| client.search.errorCode == 12 partial results*/) {
	  lsp.dbg.println(DBG.SHORT, "Search successful");
	  addRecords(1, client.search.Present);
	} else {	
	  throw new SDLIPException(SDLIPException.SERVER_ERROR_EXC, "Search failed: Z39.50 error code = " + client.search.errorCode +
				   ", message = " + client.search.errorMsg);
	}
      }
    }

    public String toString() {
      return "[SID:" + SID + "]";
    }
  }

  static final int DEFAULT_Z3950_PORT = 210;
  static final int MAX_STATE_TIMEOUT = 10 * 60; // 10 min
  static final int MAX_DOCS_PER_FETCH = 100;

  Hashtable sessions = new Hashtable(); // Integer -> Session
  int lastSessionID = 0;
  DBG dbg = new DBG();

  String host;
  int port;

  XMLObject subcolInfo;
  String defaultSubcol;

  String preferredRecordSyntax = Z39presentApi.MARC_SYNTAX;
  int queryType = 1;

  public Z39LSP(String hostPort) throws SDLIPException {
    try {
      // extract host and port, misuse URL for that
      URL lspURI = new URL("http://" + hostPort);
      host = lspURI.getHost();
      port = lspURI.getPort();
      if(port <= 0)
	port = DEFAULT_Z3950_PORT;
    } catch (MalformedURLException mue) {
      throw new SDLIPException(SDLIPException.SERVER_ERROR_EXC, "Invalid Z39.50 server access point: " + hostPort);
    }
  }

  public void setDefaultSubcol(String subcol) {
    defaultSubcol = subcol;
  }

  public void setDBG(DBG dbg) {
    this.dbg = dbg;
  }

  public void search(
		     int clientSID,
		     sdlip.XMLObject subcollections,
		     sdlip.XMLObject query,
		     int numDocs,
		     sdlip.XMLObject docProps,
		     int stateTimeoutReq,
		     sdlip.XMLObject queryOptions,
		     IntHolder expectedTotal,
		     IntHolder stateTimeout,
		     IntHolder serverSID,
		     sdlip.XMLObject serverDelegate,
		     sdlip.XMLObject result
		     )
    throws SDLIPException {


    /* get Elements from anonymous XMLObjects */
    //      Element eSubcols = XMLObject.create(subcols).getElement();
    Element eQuery = XMLObject.create(query).getElement();
    Element eDocProps = XMLObject.create(docProps).getElement();
    //      Element eQueryOptions = XMLObject.create(queryOptions).getElement();
    
    Element eWhere = DOMUtil.getChild(eQuery, DAV.where);
    Element queryBody = QueryUtil.inside(eWhere);
    
    String rpnQuery = null;
    
    // case 1: simple "contains" query
    
    String containsStr = DOMUtil.getDescendantText(eWhere, DAV.contains);
    if(containsStr != null) {
      rpnQuery =
	enquote(containsStr) + "/u=4 " +
	enquote(containsStr) + "/u=62 OR " +
	enquote(containsStr) + "/u=1003 OR";
    } else {
      // complex query
      queryBody = QueryUtil.canonicalDNF(queryBody);
      if(QueryUtil.isTrue(queryBody))
	throw new SDLIPException(SDLIPException.BAD_QUERY_EXC, "Query is always true!");
      if(QueryUtil.isFalse(queryBody)) {
	// empty set
	expectedTotal.value = 0;
	result.setString("<SearchResult xmlns='" + SDLIP.Namespace + "'/>");
	return;
      }
      rpnQuery = getRPNQuery(queryBody);
    }
    
    System.out.println("--- RPN query: " + rpnQuery);

    // create a new session
    Session session = new Session(++lastSessionID, rpnQuery, Math.min(stateTimeoutReq, MAX_STATE_TIMEOUT), this);

    if(session.isStateful()) {
      // stateful request, save the new session
      addSession(session);
    }
    String docSelection = (numDocs == sdlip.Search.RETURN_ALL_DOCS) ? "1-" : "1-" + numDocs;

    /* deliver the result */
    XMLObject.setElement(result, session.getDocs(docSelection, docProps));
    // set expected total number of results
    expectedTotal.value = session.getNumDocs();
    // set timeout after request processing for more accurate result
    if(session.isStateful()) {
      stateTimeout.value = session.getTimer().getTimeout();
      serverSID.value = session.getSID();
    } else
      session.cancel();
  }

  String getRPNQuery(Element op) throws SDLIPException {

    String orSubQuery = null;

    NodeList orList = DOMUtil.getChildElements(op, "*");
    for(int i=0; i < orList.getLength(); i++) {
      NodeList andList = DOMUtil.getChildElements((Element)orList.item(i), "*");
      String andSubQuery = null;
      // separate negated and non-negated operands
      Vector nonneg = new Vector();
      Vector neg = new Vector();
      for(int j=0; j < andList.getLength(); j++) {
	Element predicate = (Element)andList.item(j);
	if(QueryUtil.isNOT(predicate))
	  neg.addElement(predicate);
	else
	  nonneg.addElement(predicate);
      }
      // if there are no positive operands, throw an error
      if(nonneg.size() == 0)
	throw new SDLIPException(SDLIPException.BAD_QUERY_EXC, "Unbound operand(s) in the query, cannot replace with AND-NOT");
      // process non-negative
      for(int k = 0; k < nonneg.size(); k++) {
	Element predicate = (Element)nonneg.elementAt(k);
	String res = getRPNPredicate(predicate);
	if(andSubQuery == null)
	  andSubQuery = res;
	else
	  andSubQuery += " " + res + " AND";
      }
      // process negative
      for(int k = 0; k < neg.size(); k++) {
	Element predicate = (Element)neg.elementAt(k);
	andSubQuery += " " + getRPNPredicate(QueryUtil.inside(predicate)) + " NOT";
      }
      if(orSubQuery == null)
	orSubQuery = andSubQuery;
      else
	orSubQuery += " " + andSubQuery + " OR";
    }

    return orSubQuery;
  }

  String getRPNPredicate(Element op) throws SDLIPException {

    Element eProp = QueryUtil.inside(DOMUtil.getChild(op, DAV.prop));
    String value = DOMUtil.getChildText(op, DAV.literal);
    if(eProp == null || value == null)
      throw new SDLIPException(SDLIPException.BAD_QUERY_EXC, "Malformed operand: " + op.getTagName());
    String tag = eProp.getTagName();
    
    String s = null;
    
    if(DublinCore.Title.equals(tag))
      s = enquote(value) + "/u=4";
    else if(DublinCore.Creator.equals(tag))
      s = enquote(value) + "/u=1003";
    else if(DublinCore.Description.equals(tag))
      s = enquote(value) + "/u=62";
    else
      throw new SDLIPException(SDLIPException.BAD_QUERY_EXC, "Unsupported operator: " + tag);
    return s;
  }

  String enquote(String s) {
    if(s.indexOf('"') >= 0)
      return s;
    else
      return "\"" + s + "\"";
  }

  /* Method of Timeoutable */

  public void timeout(Object handle) {
    
    Session session = (Session)handle;
    dbg.println(DBG.SHORT, "Session " + session + " timed out!");
    removeSession(session);
  }

  /* Methods of the source metadata interface */


  public void getInterface(sdlip.XMLObject intf) throws SDLIPException {
    
    collectMetadata();
    intf.setString("<SDLIPInterface xmlns='" + SDLIP.Namespace + "'>" +
		   "  <SearchInterface/>" +
		   "  <ResultAccessInterface/>" +
		   (subcolInfo != null ? "  <MetadataInterface/>" : "") +
		   "</SDLIPInterface>");
  }

  public void getSubcollectionInfo(sdlip.XMLObject subcolInfo) throws SDLIPException {
    
    collectMetadata();
    if(this.subcolInfo == null)
      throw new SDLIPException(SDLIPException.NOT_IMPLEMENTED_EXC, "");
    else
      XMLObject.setElement(subcolInfo, this.subcolInfo.getElement());
  }

  public void setSubcollectionInfo(sdlip.XMLObject subcolInfo) throws SDLIPException {

    this.subcolInfo = XMLObject.create(subcolInfo);
  }


  boolean isMetadataCollected = false;

  synchronized void collectMetadata() throws SDLIPException {

    if(subcolInfo != null || isMetadataCollected)
      return;
    isMetadataCollected = true;
    
    Session session = new Session(0, null, 0, this);
    session.initialize();
    session.cancel();
  }

  public void getPropertyInfo(String subcolName, 
			      sdlip.XMLObject propInfo) throws SDLIPException {

      propInfo.setString("<propList xmlns='" + SDLIP.Namespace + "'" + 
			 "          xmlns:dc='" + DublinCore.Namespace + "'>" +
			 "  <dc:Title><searchable/><retrievable/></dc:Title>" +
			 "  <dc:Creator><searchable/><retrievable/></dc:Creator>" +
			 "  <dc:Description><searchable/><retrievable/></dc:Description>" +
			 "</propList>");
  }

  // Methods of the ResultAccess interface

  // ResultAccess implementation

  public void cancelRequest(
		     int serverSID,
		     int reqID
		     )
    throws SDLIPException {

    // we can only cancel the whole session ;(
    Session s = findSession(serverSID);
    dbg.println(DBG.SHORT, "Cancelling session " + s);
    s.cancel();
    if(reqID == ResultAccess.CANCEL_ALL_REQUESTS) {
      removeSession(s);
    }
  }

  public void getSessionInfo(
			     int serverSID,
			     IntHolder expectedTotal,
			     IntHolder stateTimeout
			     )
    throws SDLIPException {
    

    Session session = findSession(serverSID);
    expectedTotal.value = session.getNumDocs();
    stateTimeout.value = session.getTimer().getTimeout();
    dbg.println(DBG.SHORT, "Info requested for " + session + ": total=" + expectedTotal.value + ", timeout=" + stateTimeout.value);
  }
  
  public void getDocs(
		      int serverSID,
		      int reqID,
		      sdlip.XMLObject docPropList,
		      String docsToGet,
		      sdlip.XMLObject res )
    throws SDLIPException {

    Session session = findSession(serverSID);
    dbg.println(DBG.SHORT, "Getting documents " + docsToGet + " for " + session);
    XMLObject.setElement(res, session.getDocs(docsToGet, docPropList));
  }
  
  public void extendStateTimeout(
				 int serverSID,
				 int additionalTime,
				 IntHolder timeAllotted )
    throws SDLIPException {

    Session session = findSession(serverSID);
    Timer timer = session.getTimer();
    // it can happen that the killer thread is under way...
    if(timer.getTimeout() == 0)
      throw new SDLIPException(SDLIPException.REQUEST_TIMEOUT_EXC, "Invalid or timed out session");
    timer.setTimeout( Math.min(timer.getTimeout() + additionalTime, MAX_STATE_TIMEOUT) );
    timeAllotted.value = timer.getTimeout();
    dbg.println(DBG.SHORT, "Extended state timeout of " + session + " for " + timeAllotted.value);
  }
  
  // Private methods

  void addSession(Session s) {
    sessions.put(new Integer(s.getSID()), s);
    dbg.println(DBG.SHORT, "New session " + s + " created");
  }
  
  void removeSession(Session s) {
    s.cancel();
    sessions.remove(new Integer(s.getSID()));
    dbg.println(DBG.SHORT, "Removing session " + s + ", " + sessions.size() + " left");
  }
  
  Session findSession(int SID) throws SDLIPException {
    Session s = (Session)sessions.get(new Integer(SID));
    if(s == null)
      throw new SDLIPException(SDLIPException.REQUEST_TIMEOUT_EXC, "Invalid or timed out session");
    return s;
  }

  public String toString() {
    return "Z39.50 LSP serving " + host + ":" + port;
  }


  public static void main(String args[]) throws Exception {
    
    /* Instantiate an SDLIP/DASL transport module
     * This also starts the server on the given port */

    if(args.length < 2) {
      System.err.println("Usage: java sdlip.examples.Z39LSP <Z39-host>[:<Z39-port>] <port> [default subcollection]");
      System.exit(1);
    }

    Z39LSP lsp = new Z39LSP(args[0]);
    if(args.length == 3)
      lsp.setDefaultSubcol(args[2]);
    DBG dbg = new DBG(DBG.SHORT);
    lsp.setDBG(dbg);

    ServerDaslTransport t1 = new ServerDaslTransport( Integer.parseInt(args[1]) );
    t1.setDBG(dbg);
    t1.addLSP(lsp);
    
    /* Initialize CORBA transport only if runtime libraries are available */
    /*
    try {
      ServerCorbaTransport t2 = new ServerCorbaTransport();
      t2.setDBG(dbg);
      t2.addLSP(lsp);
    } catch (NoClassDefFoundError err) {}
    */
  }

}

