package sdlip.examples.demo;

import sdlip.SDLIPException;
/* holder for integer values */
import org.omg.CORBA.IntHolder;
/* transport module libraries */
import sdlip.helpers.*;
/* following packages are imported only if DOM-aware XMLObjects are used */
import sdlip.xml.dom.*;
import org.w3c.dom.*;
/* SDLIP XML element names */
import sdlip.SDLIP;
import sdlip.DAV;

import netscape.javascript.*;

import java.util.*;

/**
 * A TreeNode representing an SDLIP result set.
 */

public class ResultNode extends BasicNode implements Timeoutable {

  int TIMEOUT_INTERVAL = 10 * 60; // 5 min for the 
  static final int RESULT_SIZE = 10;

  static final int TIMER_WARN_INTERVAL = 60; // 60 seconds

  Timer timer;
  int expectedTotal;
  int serverSID;
  String server;
  CollectionNode collection;
  Element ePropList; // original properties to retrieve

  int currentItem = 1;
  Vector result = new Vector(); // 1 -> propList

  static int resultNumber = 0;
  static int requestID = 0;

  String resultName;

  int itemsPerPage = 3;

  ClientTransportModule tm;

  public ResultNode(CollectionNode coll, String name, Element eBasicsearch, Element ePropList) throws SDLIPException {

    super();
    this.resultName = (name == null || "".equals(name)) ? "Result " + (++resultNumber) : name;
    this.label = "<a href='javascript:parent.frmTop.document.clientApp.showResult(" + this.hashCode() + ")'><font color=black>" + resultName + "</font></a>";
    this.ePropList = ePropList;
    this.collection = coll;
    this.server = collection.getServer().getHost();
    this.tm = ClientTransportModule.create(server);
    tm.setDBG(new DBG(DBG.VERBOSE));
    
    String clientSID = "12345";
    //    String numDocs = !"".equals(limit) ? limit : "10";
    
    /* the result will be held in a DOM-enabled sdlip.xml.dom.XMLObject */
    XMLObject result = new XMLObject();
    IntHolder hTotal = new IntHolder(RESULT_SIZE);
    IntHolder hTimeout = new IntHolder();
    IntHolder hServerSID = new IntHolder();
    XMLObject delegate = new XMLObject();

    // subcols
    Document doc = XMLObject.getDocument();
    Element eSubcols = doc.createElement(SDLIP.subcols);
    eSubcols.appendChild(DOMUtil.createTextNode(doc, SDLIP.subcolName, collection.getName()));
    
    tm.search(Integer.parseInt(clientSID),  /* some client SID */
	      new XMLObject(eSubcols),            /* search the parent subcollection */
	      new XMLObject(eBasicsearch),
	      RESULT_SIZE,
	      new XMLObject(ePropList),            /* properties to retrieve */
	      TIMEOUT_INTERVAL, /* maintain state */
    	      null,            /* no query options */
    	      hTotal,          /* expected total number of results */
    	      hTimeout,        /* place to return stateTimeout */
    	      hServerSID,      /* place to return serverSID - don't care */
    	      delegate,        /* place to return serverDelegate - don't care */
    	      result           /* place to return search result */
    	      );
    
    //    printResults(b, result);
    if(delegate.getElement() != null) {
      // update server
      NodeList colList = DOMUtil.getChildElements(delegate.getElement(), SDLIP.serverDelegate);
      for(int i=0; i < colList.getLength(); i++) {
	Element e = (Element)colList.item(i);
	String tmp = DOMUtil.getText(e);
	if(tmp != null && tmp.startsWith("http:")) {
	  server = tmp;
	  tm = ClientTransportModule.create(server);
	}
      }
    }

    mergeResults(result.getElement());

    timer = (hTimeout.value > 0) ? new Timer(hTimeout.value - TIMER_WARN_INTERVAL, this, null) : null;
    expectedTotal = hTotal.value;
    serverSID = hServerSID.value;
    
    //    display("results", b.toString());
  }

  Element getResultItem(int i) {

    return (i >= 1 && i <= result.size()) ? (Element)result.elementAt(i - 1) : null;
  }

  // assumption: sequential results
  void mergeResults(Element e) {

    NodeList colList = DOMUtil.getChildElements(e, "*");
    for(int i=0; i < colList.getLength(); i++) {
      try {
	Element eDoc = (Element)colList.item(i);
	int num = Integer.parseInt(DOMUtil.getChildText(eDoc, SDLIP.DID));
	Element ePropList = DOMUtil.getChild(eDoc, SDLIP.propList);
	if(num > result.size())
	  result.addElement(ePropList);
	else {
	  // more sophisticated attribute-wise merging neeeded
	  mergeSubelements((Element)result.elementAt(num-1), ePropList);
	}
      } catch (NumberFormatException any) {} // should not happen
    }
  }

  void mergeSubelements(Element e1, Element e2) {

    NodeList list = DOMUtil.getChildElements(e2, "*");
    for(int i=0; i < list.getLength(); i++) {
      Element c2 = (Element)list.item(i);
      Element c1 = DOMUtil.getChild(e1, c2.getTagName());
      if(c1 == null)
	e1.appendChild(c2);
      else
	mergeSubelements(c1, c2);
    }
  }

  public CollectionNode getCollection() {
    return collection;
  }

  // from Timeoutable
  public void timeout(Object handle) {

    System.out.println("Result: " + this + " timed out");
    DemoClient client = collection.getServer().getClient();

    if(isEmpty()) { // no reason to extend the session
      return;
    }

    boolean ex = client.confirm("Result \"" + resultName + "\" in collection \"" + collection.getName() + "\" is about to expire. Extend it?");
    if(ex) {

      try {
	netscape.security.PrivilegeManager.enablePrivilege(client.PRIVILEGE);
      
	IntHolder hTimeAllotted = new IntHolder();
    
	tm.extendStateTimeout(serverSID,
			      TIMEOUT_INTERVAL,
			      hTimeAllotted);
    
	timer = hTimeAllotted.value > 0 ? new Timer(hTimeAllotted.value - TIMER_WARN_INTERVAL, this, null) : null;
        if(hTimeAllotted.value > 0) { // exponential backoff
	  TIMEOUT_INTERVAL = hTimeAllotted.value * 3;
	}
	System.out.println("Additional time allotted: " + hTimeAllotted.value);

      } catch (Exception any) {
	client.alert("Could not extend session: " + any);
      }
    }
  }

  boolean isActiveSession() {

    return timer != null && !timer.isTimedOut();
  }

  boolean canFetchMore() {
    return
      (result.size() < expectedTotal ||
       expectedTotal == sdlip.Search.UNKNOWABLE ||
       expectedTotal == sdlip.Search.NOT_YET_KNOWN) &&
      isActiveSession();
  }

  void putButton(StringBuffer b, String text, boolean enabled, int item, String bgcolor) {

    putScriptButton(b, text, enabled, "nextResultItem(" + hashCode() + ", " + item + ")", bgcolor);
  }

  void putScriptButton(StringBuffer b, String text, boolean enabled, String script, String bgcolor) {

    //    String color = enabled ? "000000" : "555555";
    b.append("<td><table border=1 cellpadding=2 cellspacing=0 bgcolor='" + bgcolor + "' width=70><tr><td width=100% align=center>");
    if(enabled)
      b.append("<a href='javascript:parent.frmTop.document.clientApp." + script + "'><font color=black><b>" + text + "</font></a>");
    else
      b.append("<font color='cccccc'><b>" + text + "</font>");
    b.append("</table></td>");
  }

  public void done() {
    if(timer != null)
      timer.interrupt();
  }

  public void displayCurrentItem(StringBuffer b) {

    b.append("<h2>Search result: <font color=darkblue>" + resultName + "</font></h2>");

    if(result.size() == 0) {
      if(!canFetchMore()) {
	b.append("<h2>No results found</h2>");
	return;
      } else { // only can fetch next results
	b.append("<a href='javascript:parent.frmTop.document.clientApp.nextResultItem(" + hashCode() + ", " + (currentItem+1) + ")'>fetch " + RESULT_SIZE + " results</a>");
	return;
      }
    }

    b.append("<form name='formIPP' onSubmit='javascript:parent.frmTop.document.clientApp.setItemsPerPage(" + hashCode() + ", document.formIPP.ipp.value); return false;'>");

    b.append("<table border=0><tr>");
    putButton(b, "first", currentItem > 1, 1, "eeffee");
    putButton(b, "previous", currentItem > 1, currentItem - itemsPerPage, "eeffee");
    putButton(b, "next", currentItem + itemsPerPage - 1 < result.size(), currentItem + itemsPerPage, "eeffee");
    putButton(b, "last", currentItem + itemsPerPage - 1 < result.size(), result.size() - itemsPerPage + 1, "eeffee");
    putScriptButton(b, "discard", true, "discardResult(" + hashCode() + ")", "ffeeee");
    b.append("</tr></table>");

    //    b.append("<h2>Item: #" + currentItem + " of " + result.size() + "/" + expectedTotal + "</h2>");
    b.append("<p>Total search results: <b>" + expectedTotalString(expectedTotal) + "</b>");
    b.append(", fetched: <b>" + result.size() + "</b>");

    if(canFetchMore()) {
      b.append("<br><a href='javascript:parent.frmTop.document.clientApp.nextResultItem(" + hashCode() + ", " + (result.size()+1) + ")'>Fetch next " + RESULT_SIZE + " results</a>");
    }

    b.append("<br>Show results per page: <input type='text' name='ipp' size=3 maxlength=3 value='" + itemsPerPage + "'>");
    b.append("</form>");

    b.append("<form name='formMore' onSubmit='return false;'>");
    b.append("<p><table border=0 cellpadding=3 cellspacing=3>");

    int limit = Math.min(currentItem + itemsPerPage - 1, result.size());
    for(int i = currentItem; i <= limit; i++) {

      b.append("<tr><td align=center><b>" + i);
      if(isActiveSession())
	b.append("<br><input type='checkbox' name='did_" + i + "'>");
      b.append("</td><td>");
      b.append("<table border=1 cellpadding=3 cellspacing=1>");
      Element item = getResultItem(i);
      for(Enumeration en = collection.getAttributes(); en.hasMoreElements();) {

	NestedAttribute a = (NestedAttribute)en.nextElement();
	for(Enumeration ven = a.matchValues(item); ven.hasMoreElements();) {
	  String value = (String)ven.nextElement();
	  //	System.out.println("Value for " + a.getPropName() + " is " + value);
	  b.append("<tr><td bgcolor='ffffaa'><small><b>" + a.getPropName());
	  b.append("</td><td bgcolor='fffffee'>");
	  String lv = value.toLowerCase();
	  if(value.startsWith("http://")) {
	    if (value.endsWith(".gif") || value.endsWith(".jpg"))
	      b.append("<img src='" + value + "'>");
	    else
	      b.append("<a href='" + value + "'>" + value + "</a>");
	  } else
	    b.append(value);
	  b.append("</td></tr>");
	}
      }
      b.append("</table>");
      b.append("</td></tr>");
    }
    b.append("</table>");

    if(isActiveSession()) {
      b.append("<a href='javascript:parent.frmTop.document.clientApp.attributeChoice(" + hashCode() + ")'>Choose additional attributes</a> : ");
      b.append("<a href='javascript:parent.frmTop.document.clientApp.retrieveMoreAttributes(" + hashCode() + ", document.formMore)'>retrieve additional attributes</a> for selected items");
    }

    b.append("</form>");

  }

  public boolean isEmpty() {
    return (expectedTotal == 0) || (result.size() == 0 && !isActiveSession());
  }

  public void retrieveMoreAttributes(JSObject form) throws Exception {

    Vector did = new Vector();
    String docsToGet = "";

    try {
      for(int i = currentItem; i < currentItem + itemsPerPage; i++) {
	JSObject check = (JSObject) form.getMember("did_" + i);
	//	System.err.println("Got: " + check);
	if(((Boolean) check.getMember("checked")).booleanValue()) {
	  did.addElement(new Integer(i));
	  //	  System.err.println("RETRIEVE FOR: " + i);
	  if(!"".equals(docsToGet))
	    docsToGet += ",";
	  docsToGet += String.valueOf(i);
	}
      }
    } catch (Exception any) {
      //      System.err.println("EXC: " + any);
    }

    if(!"".equals(docsToGet)) {
      
      XMLObject r = new XMLObject();
      tm.getDocs(serverSID,
		 ++requestID,
		 new XMLObject(collection.getPropListToRetrieve()),
		 docsToGet,
		 r);
      mergeResults(r.getElement());
    }
  }

  public void setItemsPerPage(int ipp) {

    itemsPerPage = ipp;
    if(currentItem + ipp - 1 > result.size())
      currentItem = result.size() - ipp + 1;
    if(currentItem < 1)
      currentItem = 1;
  }


  public void nextResult(int num) throws SDLIPException {

    if(num < 1)
      num = 1;

    while(num > result.size() && canFetchMore()) {
      
      XMLObject r = new XMLObject();
      
      String docsToGet = "" + num + "-" + (num + RESULT_SIZE - 1);
      tm.getDocs(serverSID,
		 ++requestID,
		 new XMLObject(ePropList),
		 docsToGet,
		 r);
      
      int oldResultNum = result.size();
      mergeResults(r.getElement());

      if(oldResultNum + RESULT_SIZE != result.size() && expectedTotal < 0) {
	// no more documents?
	IntHolder hExpectedTotal = new IntHolder();
	IntHolder hStateTimeout = new IntHolder();

	tm.getSessionInfo(serverSID, hExpectedTotal, hStateTimeout);
	expectedTotal = hExpectedTotal.value;
	if(expectedTotal < 0) { // give up
	  expectedTotal = result.size();
	}
      }
    } // while
    if(num <= result.size())
      currentItem = num;
    else
      num = result.size();
  }

  public static String expectedTotalString(int total) {
    
    if(total == sdlip.Search.UNKNOWABLE)
      return "unknowable";
    else if(total == sdlip.Search.NOT_YET_KNOWN)
      return "not yet known";
    else
      return String.valueOf(total);
  }


}
