package sdlip.helpers;

import sdlip.Search;
import sdlip.ResultAccess;
import sdlip.Metadata;
import sdlip.SDLIPException;
import sdlip.SDLIP;

import java.util.Hashtable;

public abstract class ClientTransportModule implements Search, ResultAccess, Metadata {

  /**
   * List of registered transport modules: (lowercase URI prefix, Class)
   */
  static Hashtable tms = new Hashtable();
  static boolean initialized = false;

  protected DBG dbg = new DBG();

  public void setDBG(DBG dbg) {
    this.dbg = dbg;
  }

  /** Default initialization. Returns true if this module accepts the URI */
  public void setLspUri(String uri) throws SDLIPException {
  }

  static void initialize() {

    // try to load well-known transport modules
    initialized = true;
    register("http", "sdlip.helpers.ClientDaslTransport");
    register(SDLIP.NameServerURISchema, "sdlip.helpers.ClientCorbaTransport");
    register(SDLIP.DienstServerURISchema, "sdlip.helpers.ClientDienstTransport");
  }

  public static final void register(String uriPrefix, String className) {
    
    tms.put(uriPrefix, className);
  }

  public static final void deregister(String uriPrefix) {

    tms.remove(uriPrefix);
  }

  public static final ClientTransportModule create(String uri) throws SDLIPException {

    if(!initialized)
      initialize();

    Class cls = null;

    int i = uri.indexOf(':');
    if(i > 0) {
      String prefix = uri.substring(0, i).toLowerCase();
      String className = (String)tms.get(prefix);
      if(className != null) {
	try {
	  cls = Class.forName(className);
	} catch (Exception any) {}
      }
    }

    if(cls == null)
      throw new RuntimeException("[Client transport module] do not know how to create transport module for " + uri);

    try {
      ClientTransportModule tm = (ClientTransportModule)cls.newInstance();
      tm.setLspUri(uri);
      return tm;
    } catch (Exception any) {
      throw new RuntimeException("[Client transport module] cannot create transport module: " + any);
    }
  }
  
}
