package sdlip.helpers;

import java.io.*;
import java.util.*;
import java.net.*;

import sdlip.CORBA.*;
import sdlip.http.*;

/**
 * Acts as a wrapper or as a transport module
 */
public class ServerCorbaTransport extends POA_GenericDispatch implements HTTPCallback {

  static org.omg.CORBA.ORB orb;
  static org.omg.PortableServer.POA poa;

  static DBG dbg = new DBG();

  Object lsp;
  String ior; // ior of the lsp

  /** lsp path -> Search */
  static Hashtable lsps = new Hashtable();

  public ServerCorbaTransport(String lspPath, Object lsp) throws Exception {
    
    this.lsp = lsp;
    lsps.put(lspPath, this);
    register(this);
  }

  public ServerCorbaTransport() {
  }

  public ServerCorbaTransport(Object lsp) throws Exception {
    this("/", lsp);
  }

  public void setDBG(DBG dbg) {

    this.dbg = dbg;
  }

  public void addLSP(String lspPath, Object lsp) throws Exception {

    if(!( (lsp instanceof sdlip.Search) ||
	  (lsp instanceof sdlip.ResultAccess)))
      throw new RuntimeException("LSP must at least implement Search or ResultAccess interface");
    ServerCorbaTransport tm = new ServerCorbaTransport(lspPath, lsp);
  }

  public void addLSP(Object lsp) throws Exception {
    
    addLSP("/", lsp);
  }

  public ServerCorbaTransport getLSP(String lspPath) {

    return (ServerCorbaTransport)lsps.get(lspPath);
  }

  protected Object getServerImpl() {
    return lsp;
  }

  String getLSPPath(String line) {

    StringTokenizer st = new StringTokenizer(line);
    st.nextToken();
    return st.nextToken();
  }

  public void search(int clientSID,
		     java.lang.String subcols,
		     java.lang.String query,
		     int numDocs,
		     java.lang.String docPropList,
		     int stateTimeoutReq,
		     java.lang.String queryOptions,
		     org.omg.CORBA.IntHolder expectedTotal,
		     org.omg.CORBA.IntHolder stateTimeout,
		     org.omg.CORBA.IntHolder serverSID,
		     org.omg.CORBA.StringHolder serverDelegate,
		     org.omg.CORBA.StringHolder result)
    throws sdlip.CORBA.SDLIPException {

    SimpleXMLObject xServerDelegate = new SimpleXMLObject();
    SimpleXMLObject xResult = new SimpleXMLObject();
    // default value for expected total
    expectedTotal.value = sdlip.Search.UNKNOWABLE;

    //    System.err.println("Invoking search");
    try {
      checkSearchInterface(lsp);
      ((sdlip.Search)lsp).search(clientSID,
				 new SimpleXMLObject(subcols),
				 new SimpleXMLObject(query),
				 numDocs,
				 new SimpleXMLObject(docPropList),
				 stateTimeoutReq,
				 new SimpleXMLObject(queryOptions),
				 expectedTotal,
				 stateTimeout,
				 serverSID,
				 xServerDelegate,
				 xResult);
    } catch (Exception se) {
      raiseCorbaExc(se);
    }
    serverDelegate.value = notNull(xServerDelegate.getString());
    result.value = notNull(xResult.getString());
    //    System.err.println("search done: " + serverDelegate.value + ", " + result.value);
  }

  public void getSessionInfo(int serverSID, org.omg.CORBA.IntHolder expectedTotal, org.omg.CORBA.IntHolder stateTimeout)
    throws sdlip.CORBA.SDLIPException {

    try {
      checkResultAccessInterface(lsp);
      ((sdlip.ResultAccess)lsp).getSessionInfo(serverSID,
					       expectedTotal,
					       stateTimeout);
    } catch (Exception se) {
      raiseCorbaExc(se);
    }
  }

  public void getDocs(int serverSID, int reqID, java.lang.String docPropList, java.lang.String docsToGet, org.omg.CORBA.StringHolder res)
    throws sdlip.CORBA.SDLIPException {

    SimpleXMLObject xResult = new SimpleXMLObject();
    try {
      checkResultAccessInterface(lsp);
      ((sdlip.ResultAccess)lsp).getDocs(serverSID,
					reqID,
					new SimpleXMLObject(docPropList),
					docsToGet,
					xResult);
    } catch (Exception se) {
      raiseCorbaExc(se);
    }
    res.value = notNull(xResult.getString());
  }

  public void extendStateTimeout(int serverSID, int additionalTime, org.omg.CORBA.IntHolder timeAllotted) throws sdlip.CORBA.SDLIPException {

    try {
      checkResultAccessInterface(lsp);
      ((sdlip.ResultAccess)lsp).extendStateTimeout(serverSID,
						   additionalTime,
						   timeAllotted);
    } catch (Exception se) {
      raiseCorbaExc(se);
    }
  }

  public void cancelRequest(int serverSID, int reqID) throws sdlip.CORBA.SDLIPException {

    try {
      checkResultAccessInterface(lsp);
      ((sdlip.ResultAccess)lsp).cancelRequest(serverSID,
					      reqID);
    } catch (Exception se) {
      raiseCorbaExc(se);
    }
  }

  public void getInterface(org.omg.CORBA.StringHolder theInterface) throws sdlip.CORBA.SDLIPException {

    SimpleXMLObject xResult = new SimpleXMLObject();
    try {
      checkMetadataInterface(lsp);
      ((sdlip.Metadata)lsp).getInterface(xResult);
    } catch (Exception se) {
      raiseCorbaExc(se);
    }
    theInterface.value = notNull(xResult.getString());
  }

  public void getSubcollectionInfo(org.omg.CORBA.StringHolder subcolInfo) throws sdlip.CORBA.SDLIPException {

    SimpleXMLObject xResult = new SimpleXMLObject();
    try {
      checkMetadataInterface(lsp);
      ((sdlip.Metadata)lsp).getSubcollectionInfo(xResult);
    } catch (Exception se) {
      raiseCorbaExc(se);
    }
    subcolInfo.value = notNull(xResult.getString());
  }

  public void getPropertyInfo(java.lang.String subcolName, org.omg.CORBA.StringHolder propInfo) throws sdlip.CORBA.SDLIPException {

    SimpleXMLObject xResult = new SimpleXMLObject();
    try {
      checkMetadataInterface(lsp);
      ((sdlip.Metadata)lsp).getPropertyInfo("".equals(subcolName) ? null : subcolName, xResult);
    } catch (Exception se) {
      raiseCorbaExc(se);
    }
    propInfo.value = notNull(xResult.getString());
  }

  String notNull(String s) {
    return s == null ? "" : s;
  }

  void raiseCorbaExc(Exception ex) throws sdlip.CORBA.SDLIPException {

    if(ex instanceof sdlip.SDLIPException) {
      sdlip.SDLIPException se = (sdlip.SDLIPException)ex;
      String details = null;
      try {
	details = se.getDetails().getString();
      } catch (Exception any) {}
      throw new sdlip.CORBA.SDLIPException(se.getCode(), notNull(se.getReason()), notNull(details));
    } else {
      dbg.println(DBG.SHORT, "Internal exception: " + ex);
      throw new sdlip.CORBA.SDLIPException(sdlip.SDLIPException.SERVICE_UNAVAILABLE_EXC, String.valueOf(ex), "");
    }
  }


  void checkSearchInterface(Object obj) throws sdlip.SDLIPException {
    if(!(obj instanceof sdlip.Search))
      throw new sdlip.SDLIPException(sdlip.SDLIPException.NOT_IMPLEMENTED_EXC, "Search interface not implemented");
  }

  void checkResultAccessInterface(Object obj) throws sdlip.SDLIPException {
    if(!(obj instanceof sdlip.ResultAccess))
      throw new sdlip.SDLIPException(sdlip.SDLIPException.NOT_IMPLEMENTED_EXC, "ResultAccess interface not implemented");
  }

  void checkMetadataInterface(Object obj) throws sdlip.SDLIPException {
    if(!(obj instanceof sdlip.Metadata))
      throw new sdlip.SDLIPException(sdlip.SDLIPException.NOT_IMPLEMENTED_EXC, "Metadata interface not implemented");
  }


  /** HTTP callback implementation */
  public void process(Socket socket, InputStream in, OutputStream out) throws Exception {


    // SM.util.FileUtils.copyStream(in, System.out);

    PrintWriter pw = new PrintWriter(new OutputStreamWriter(out));

    try {

      BufferedReader din = new BufferedReader(new InputStreamReader( in ));
	  
      String line = din.readLine();

      if(line.startsWith("GET ")) {
	
	String lspPath = getLSPPath(line);
	ServerCorbaTransport lsp = getLSP(lspPath);
	if(lsp == null) {
	  dbg.println(DBG.SHORT, toString() + " no LSP for " + lspPath + " (requested by " +  socket.getInetAddress().getHostName() + ")");
	  pw.println("HTTP/1.0 400 Not Found\nContent-Type: text/html");
	  pw.println();
	  pw.println("<p>No CORBA object found for \"" + lspPath + "\".");
	  if(lsps.size() > 0) {
	    pw.println("<p>Available LSPs:<p><table cellpadding=5 border=1><tr><td><b>LSP</td><td><b>Description</td></tr>");
	    for(Enumeration en = lsps.keys(); en.hasMoreElements();) {
	      String lp = (String)en.nextElement();
	      pw.println("<tr><td><tt><a href='" + lp + "'>" + lp + "</a></tt></td><td>" + getLSP(lp).getServerImpl() + "</td></tr>");
	    }
	    pw.println("</table>");
	  } else
	    pw.println("<p>No LSPs are registered.");
	  pw.close();
	  return;
	}
	dbg.println(DBG.SHORT, toString() + " LSP for " + lspPath + " is " + lsp.getServerImpl() + " (requested by " +  socket.getInetAddress().getHostName() + ")");
	pw.println("HTTP/1.0 200 OK\nContent-Type: text/plain");
	pw.println();
	pw.println(lsp.ior);
	pw.println();
	pw.close();
      }
    } catch (Exception any) {
      dbg.println(any);
    }
  }

  void register(org.omg.PortableServer.Servant servant) throws Exception {

    boolean initORB = (orb == null);

    //    System.err.println("---REG: " + initORB);
    
    if(initORB) {
      // check IntHolder to see whether the correct class is being used
      if(!((new org.omg.CORBA.IntHolder()) instanceof org.omg.CORBA.portable.Streamable))
	throw new RuntimeException("The SDLIP/CORBA transport module requires that the org.omg.CORBA package appears in your CLASSPATH before the SDLIP package. Make sure corba-sdlip.jar goes in front of sdlip.jar.");
      orb = org.omg.CORBA.ORB.init(); // args, null);
      poa = org.omg.PortableServer.POAHelper.narrow(orb.resolve_initial_references("RootPOA"));
    }

    org.omg.CORBA.Object o = poa.servant_to_reference(servant);
    
    this.ior = orb.object_to_string( o );

    if(initORB) {
      //      orb.run();
      // start Web Server
      int port = sdlip.SDLIP.NameServerDefaultPort;
      try {
	port = Integer.parseInt( System.getProperty(sdlip.SDLIP.NameServerJVMParName) );
      } catch (Exception any) {}
      
      dbg.println(DBG.SHORT, "Starting HTTP-based CORBA name server on port: " + port);
      new HTTPServer(this, port);
    }
  }

  public String toString() {
    return "[SDLIP/CORBA Server Transport]";
  }
}

