/**
 * Copyright  Sergey Melnik (Stanford University, Database Group)
 *
 * All Rights Reserved.
 */

package sdlip.http;

import java.io.*;
import java.net.*;
import java.util.*;

/**
 * A new Connection object is created for each new connection
 * established between the server and a client.
 *
 * @author Sergey Melnik <melnik@db.stanford.edu>
 */

public class HTTPConnection extends Thread {
  
  HTTPServer       http;
  Socket          socket;
  InputStream     in;
  OutputStream    out;
  
  // The communication streams are initiated and the thread is started.
  HTTPConnection(HTTPServer http, Socket socket, ThreadGroup threadgroup, int priority) {

    // The Connection thread is given a group, a name, and a priority.
    super(threadgroup, "HTTP Connection");
    
    //      System.out.println("New connection from " + client_socket);
    
    this.http = http;
    this.setPriority(priority);
    this.socket = socket;

    // The data-exchange streams are created.
    try {
      in = socket.getInputStream();
      out = socket.getOutputStream();
      //      in = new BufferedInputStream(socket.getInputStream());
      //      out = new BufferedOutputStream(socket.getOutputStream());
    } catch (IOException e) {
      try {
	socket.close();
      } catch (IOException e2) { ; }
	
      System.err.println(http + " Exception while getting socket streams: " + e);
      return;
    }
    // The Connection thread is started.
    this.start();
  }

  public InputStream getInputStream() {
    return in;
  }

  public OutputStream getOutputStream() {
    return out;
  }

  public void run() {
    
    try {
      
      http.getCallback().process(socket, in, out);
      
    } catch (Exception any) {
      System.err.println(toString() + " ERROR: " + any);
      any.printStackTrace(System.err);
    } finally {
      try {
	//	System.err.println(http + " Hanging up connection.");
	socket.close();
      } catch (IOException sock) {}
    }
  }

}

