package sdlip.xml.dom;

import org.w3c.dom.*;

public abstract class CharacterDataImpl
    extends NodeImpl 
    implements CharacterData {

    // Constructors
    protected CharacterDataImpl(DocumentImpl ownerDocument, String data) {
        super(ownerDocument, null, data);
    }

    // Node methods
    public abstract String getNodeName();

    // CharacterData methods
    public String getData() {
        return value;
    }

    public int getLength() {   
        return value.length();
    }  

    public void appendData(String data) {

        value += data;

    }

    public void deleteData(int offset, int count) 
        throws DOMException {

        if (count < 0) {
        	throw new DOMExceptionImpl(DOMException.INDEX_SIZE_ERR, null);
        }

        int tailLength = Math.max(value.length() - count - offset, 0);
        try {
            value = value.substring(0, offset) + (tailLength > 0 ? value.substring(offset + count, offset + count + tailLength) : "");
        }
        catch (StringIndexOutOfBoundsException e) {
        	throw new DOMExceptionImpl(DOMException.INDEX_SIZE_ERR, null);
        }

    }

    public void insertData(int offset, String data) 
        throws DOMException {

        try {
        	value = new StringBuffer(value).insert(offset, data).toString();
        }
        catch (StringIndexOutOfBoundsException e) {
        	throw new DOMExceptionImpl(DOMException.INDEX_SIZE_ERR, null);
        }

    }

    public void replaceData(int offset, int count, String data) 
        throws DOMException {

        deleteData(offset, count);
        insertData(offset, data);

    }

    public void setData(String value) 
        throws DOMException {
        setNodeValue(value);
    }

    public String substringData(int offset, int count) 
        throws DOMException {

        int length = value.length();
        if (count < 0 || offset < 0 || offset > length - 1) {
            throw new DOMExceptionImpl(DOMException.INDEX_SIZE_ERR, null);
        }

        int tailIndex = Math.min(offset + count, length);

        return value.substring(offset, tailIndex);

    }

}
