package sdlip.xml.dom;

import org.w3c.dom.*;

import org.xml.sax.*;
import org.xml.sax.helpers.*;

import java.util.*;

/**
 */
public class DOMParser implements EntityResolver, DTDHandler, DocumentHandler, ErrorHandler {

  static String DEFAULT_PARSER = "com.microstar.xml.SAXDriver";

  DocumentImpl document; // used as a factory only! getDocumentElement is always null...
  ElementImpl root;

  // Constructors
  
  /** Default constructor. */
  public DOMParser() {
  }
  
  /** used by XMLObjectImpl */
  protected DOMParser(DocumentImpl d) {
    
    this.document = d;
  }
  
  public Element parse(InputSource source) throws Exception {

    return parse(source, DEFAULT_PARSER);
  }

    /** Parses the specified URI and returns the document. */
  public Element parse(InputSource source, String parserClass) throws Exception {
      
    Parser p = ParserFactory.makeParser(parserClass);
    
    // Register the handlers
    p.setEntityResolver(this);
    p.setDTDHandler (this);
    p.setDocumentHandler(this);
    p.setErrorHandler (this);
    p.parse(source);

    root.normalize();
    document.setDocumentElement(root); // FIXME: always?
    return root;
  } // parse(String):Document

  // ErrorHandler methods
  
  /** Warning. */
  public void warning(SAXParseException ex) {
    //        System.err.println("[Warning] "+
    //                           getLocationString(ex)+": "+
    //                           ex.getMessage());
  }
  
  /** Error. */
  public void error(SAXParseException ex) {
    //        System.err.println("[Error] "+
    //                           getLocationString(ex)+": "+
    //                           ex.getMessage());
  }
  
  /** Fatal error. */
  public void fatalError(SAXParseException ex) throws SAXException {
    
    //      System.err.println("[Fatal Error] "+
    //			 getLocationString(ex)+": "+
    //			 ex.getMessage());
    throw ex;
  }

  /** Returns a string of the location. */
  public static String getLocationString(SAXParseException ex) {
    StringBuffer str = new StringBuffer();
    
    String systemId = ex.getSystemId();
    if (systemId != null) {
      int index = systemId.lastIndexOf('/');
      if (index != -1) 
	systemId = systemId.substring(index + 1);
      str.append(systemId);
    }
    str.append(':');
    str.append(ex.getLineNumber());
    str.append(':');
    str.append(ex.getColumnNumber());
    
    return str.toString();
    
  } // getLocationString(SAXParseException):String
  

  // Implementation of SAX interfaces

  public InputSource resolveEntity (String publicId, String systemId) {
    return null;
  }
  
  public void notationDecl (String name, String publicId, String systemId) {
  }
  
  public void unparsedEntityDecl (String name,
				  String publicId,
				  String systemId,
				  String notationName) {
  }
  
  public void setDocumentLocator (Locator locator) {
  }
  
  public void startDocument () {
    if(document == null)
      document = new DocumentImpl();
  }
  
  public void endDocument () throws SAXException {
  }

  public void doctype (String name, String publicID, String systemID) {
  }

  /**
   * Called for each new element.
   * Build up the document tree using an element stack
   *
   * @exception SAXException Passed on since we don't handle it.
   */

  public void startElement (String name, AttributeList al) throws SAXException {

    Hashtable namespaces = HT_EMPTY; // assume none
    
    int iLength = al.getLength ();
    for (int x = 0; x < iLength; x++) {

      String aName = al.getName (x);

      if (aName.equals (XMLNS)) {

	String aValue = al.getValue (aName);
	if (aValue != null && aValue.length() == 0 && source != null)
	  aValue = source;
	namespaces = addNamespace(namespaces, aName, aValue);

      } else if (aName.startsWith (XMLNS + ":")) {

	String aValue = al.getValue (aName);
	if (aValue != null && aValue.length() == 0 && source != null)
	  aValue = source;
	aName = aName.substring (6);
	namespaces = addNamespace(namespaces, aName, aValue);
	
      }
    }

    /**
     * Place new namespace declarations into the stack
     */
    m_namespaceStack.push (namespaces);

    /**
     * Figure out the prefix part if it exists and
     * determine the namespace of the element accordingly
     */
    String sNamespace = null;
    String sElementName = null;
    int i = name.indexOf (':');
    String sPrefix2 = null;
    if (i > 0) {
      sPrefix2 = name.substring (0, i);
      sNamespace = namespace (sPrefix2);
      sElementName = name.substring (i+1);
    } else {
      sNamespace = namespace (XMLNS);
      sElementName = name;
    }

    document.registerNamespace(sNamespace);
    ElementImpl element = new ElementImpl( document, sNamespace + sElementName );

    //    System.out.println("--- ELEM: " + element.getNodeName());

    /**
     * Finally look for attributes other than the special xmlns,
     * expand them, and place to the new AttributeListImpl
     */
    for (int x = 0; x < iLength; x++) {
      String sAttributeNamespace = null;
      String aName = al.getName (x);

      if (!aName.startsWith (XMLNS)) {
	String aValue = al.getValue (aName);
	//	String aType = al.getType (aName);

	int iIndex = aName.indexOf (':');
	String sPrefix = null;
	if (iIndex > 0) {
	  sPrefix = aName.substring (0, iIndex);
	  sAttributeNamespace = namespace (sPrefix);
	  aName = aName.substring (iIndex+1);
	  
	} else {
	  if (sNamespace == null)
	    sAttributeNamespace = namespace (XMLNS);
	  else 
	    sAttributeNamespace = sNamespace;
	}

	document.registerNamespace(sAttributeNamespace);

	// add to wrapper
	//	elemWrap.setAttributeNode( new AttrWrapper(attr, sAttributeNamespace + aName) );
	//	System.out.println("--- ATTR: " + sAttributeNamespace+aName + "=" + aValue);
	element.setAttribute( sAttributeNamespace+aName,  aValue);
      }
    }

    if(root == null)
      root = element;
    
    /**
     * Update the containment hierarchy
     * with the stack.
     */
    if (!m_elementStack.empty()) {
      Element e = (Element)m_elementStack.peek ();
      e.appendChild (element);
      //      System.err.println("--- APP " + element + " to " + e);
    }

    /**
     * Place the new element into the stack
     */
    m_elementStack.push (element);

  }


  /** 
   * For each end of an element scope step back in the 
   * element and namespace stack
   *
   * @exception SAXException Passed on since we don't handle it.
   */
  public void endElement (String name) throws SAXException {
    
    m_elementStack.pop();
    // BUG fixed 10/07/00 by SM, reported by Noah Green, Columbia
    m_namespaceStack.pop();
  }


  public void characters (char ch[], int start, int length)
    throws SAXException {
      
      /**
       * Place all characters as Data instance to the containment
       * hierarchy with the help of the stack.
       */
      Element e = (Element)m_elementStack.peek ();
      String s = new String (ch, start, length);
      
      e.appendChild (document.createTextNode(s));
  }

  public void ignorableWhitespace (char ch[], int start, int length) {
  }
  
  public void processingInstruction (String target, String data) {
  }


  static final String XMLNS = "xmlns";
  static final String XMLSCHEMA = "xml";
    /**
     * Return the full namespace URI for a given prefix <i>sPrefix</i>.
     * The default namespace is identified with <i>xmlns</i> prefix.
     * The namespace of <i>xmlns</i> attribute is an empty string.
     */

    private Stack		m_namespaceStack = new Stack ();
    private Stack		m_elementStack = new Stack ();

  private String namespace (String sPrefix) {
    if (sPrefix == null) {
      sPrefix = XMLNS;
    }
    for (int x = m_namespaceStack.size()-1; x >=0; x--) {
      Hashtable ht = (Hashtable)m_namespaceStack.elementAt (x);
      String sURI = (String)ht.get (sPrefix);
      if (sURI != null)
	return sURI;
    }
    /**
     * Give error only if
     * 1. the prefix is not from the reserved xml namespace
     * 2. the prefix is not xmlns which is to look for the default
     *    namespace
     */
    if (sPrefix.equals (XMLSCHEMA)) {
      return XMLSCHEMA;
    } else if (sPrefix.equals (XMLNS)) {
      return "";
    } else {
      System.err.println ("Unresolved namespace prefix "+sPrefix);
    }
    return "";
  }

  static final Hashtable HT_EMPTY = new Hashtable();

  Hashtable addNamespace(Hashtable namespaces, String name, String value) {

    if(namespaces == HT_EMPTY)
      namespaces = new Hashtable();
    namespaces.put(name, value);
    return namespaces;
  }

  // to save space we reuse strings
  Hashtable names = new Hashtable();
  String source = null; // source location unknown


}
