package sdlip.xml.dom;

import java.util.Enumeration;
import java.util.Vector;

import org.w3c.dom.*;

public class ElementImpl 
    extends NodeImpl 
    implements Element {

    // Data

    /** Attributes. */
    protected NamedNodeMapImpl attributes;

    // Constructors

    /** Factory constructor. */
    public ElementImpl(DocumentImpl ownerDoc, String name) {
    	super(ownerDoc, name, null);
    	attributes = new NamedNodeMapImpl(ownerDoc);
    }

    // Node methods

    public short getNodeType() {
        return Node.ELEMENT_NODE;
    }

    /** Returns the node value. */
    public String getNodeValue() {
        return null;
    }

    public void setNodeValue(String value) throws DOMException {
    	throw new DOMExceptionImpl(DOMException.NO_MODIFICATION_ALLOWED_ERR, null);
    }

    public NamedNodeMap getAttributes() {
        return attributes;
    }  

    public Node cloneNode(boolean deep) {
    	ElementImpl newnode = (ElementImpl) super.cloneNode(deep);
    	// Replicate NamedNodeMap rather than sharing it.
    	newnode.attributes = attributes.cloneMap();
    	return newnode;
    }

    // Element methods

    /** Returns the element value. */
    public String getValue() {
        return null;
    }

    public String getAttribute(String name) {
        Attr attr=(Attr)(attributes.getNamedItem(name));
        return (attr==null) ? "" : attr.getValue();
    }    

    public Attr getAttributeNode(String name) {
        return (Attr)(attributes.getNamedItem(name));
    }    

    public NodeList getElementsByTagName(String tagname) {
      return new DeepNodeListImpl(this,tagname);
    }  

    public String getTagName() {
    	return name;
    }  

    public void normalize() {

    	Node kid, next;
    	for (kid = getFirstChild(); kid != null; kid = next) {
    		next = kid.getNextSibling();

    		// If kid and next are both Text nodes (but _not_ CDATASection,
    		// which is a subclass of Text), they can be merged.
    		if (next != null
			 && kid.getNodeType() == Node.TEXT_NODE 
			 && next.getNodeType() == Node.TEXT_NODE) 
    	    {
	      //    			((Text)kid).appendData(next.getNodeValue());
	      kid.setNodeValue( kid.getNodeValue() + next.getNodeValue() ); // slow but no text
    			removeChild(next);
    			next = kid; // Don't advance; there might be another.
    		}

    		// Otherwise it might be an Element, which is handled recursively  
    		else if (kid.getNodeType() ==  Node.ELEMENT_NODE) {
                ((Element)kid).normalize();
            }
        }

    	// changed() will have occurred when the removeChild() was done,
    	// so does not have to be reissued.

        } // normalize()

    public void removeAttribute(String name) {

    	AttrImpl att = (AttrImpl) attributes.getNamedItem(name);
    	// Remove it (and let the NamedNodeMap recreate the default, if any)
    	if (att != null) {
	  //    		att.owned = false;
    		attributes.removeNamedItem(name);
    	}

    } // removeAttribute(String)

    public Attr removeAttributeNode(Attr oldAttr) 
        throws DOMException 
        {

    	AttrImpl found = (AttrImpl) attributes.getNamedItem(oldAttr.getName());

    	// If it is in fact the right object, remove it (and let the
    	// NamedNodeMap recreate the default, if any)

    	if (found == oldAttr) {
    		attributes.removeNamedItem(oldAttr.getName());
		//    		found.owned = false;
    		return found;
    	}

        throw new DOMExceptionImpl(DOMException.NOT_FOUND_ERR, null);

    } // removeAttributeNode(Attr):Attr

    public void setAttribute(String name, String value) {

    	AttrImpl newAttr = (AttrImpl)getOwnerDocument().createAttribute(name);
    	newAttr.setNodeValue(value);
    	attributes.setNamedItem(newAttr);

    } // setAttribute(String,String)

    public Attr setAttributeNode(Attr newAttr) 
        throws DOMException
        {

    	if (!(newAttr instanceof AttrImpl)) {
    		throw new DOMExceptionImpl(DOMException.WRONG_DOCUMENT_ERR, null);
        }

    	AttrImpl na = (AttrImpl) newAttr;
    	AttrImpl oldAttr = (AttrImpl) attributes.getNamedItem(newAttr.getName());

    	// This will throw INUSE if necessary
    	attributes.setNamedItem(na);
	//    	na.owned = true; // Must set after adding ... or within?

    	return oldAttr;

    } // setAttributeNode(Attr):Attr

} // class ElementImpl
