package sdlip.xml.dom;

import sdlip.SDLIPException;
import org.w3c.dom.*;
import org.xml.sax.*;
import java.io.*;

/**
 * This is a DOM-enabled implementation of the SDLIP XMLObject interface
 * described in the SDLIP Java tutorial.
 */

public class XMLObject implements sdlip.XMLObject {

  static DocumentImpl document = new DocumentImpl(); // all elements are created within this document

  Element root;

  /**
   * Creates an empty XMLObject
   */
  public XMLObject() {
  }

  public XMLObject(Element e) {

    setElement(e);
  }

  public XMLObject(String xmldata) throws sdlip.SDLIPException {

    setString(xmldata);
  }

  public XMLObject(InputSource source) throws sdlip.SDLIPException {

    setInputSource(source);
  }

  public static Document getDocument() {

    return document;
  }

  public void setElement(Element e) {

    root = e;
  }

  public Element getElement() {
    return root;
  }
  
  /**
   * Standard get method
   */
  public String getString() {
    
    return root == null ? null : DOMWriter.serialize(root);
  }

  /**
   * WARNING: Use this for debugging purposes only.
   * Whitespace is not preserved in the generated output!
   */
  public String getPrettyString() {

    return DOMWriter.prettyPrint(root);
  }
  
  /**
   * Standard set method
   */
  public void setString( String XMLStr ) throws SDLIPException {
    
    if(XMLStr != null && XMLStr.length() > 0)
      setInputSource( new InputSource(new StringReader(XMLStr)) );
  }

  public void setInputSource( InputSource source ) throws SDLIPException {
    
    try {
      DOMParser p = new DOMParser(document);
      root = p.parse(source);
    } catch(SAXParseException pe) {
      throw new SDLIPException(SDLIPException.MALFORMED_XML_EXC, DOMParser.getLocationString(pe)+": "+ pe.getMessage());
    } catch(SAXException sax) {
      String msg = sax.getException() != null ? sax.getException().getMessage() : sax.getMessage();
      throw new SDLIPException(SDLIPException.MALFORMED_XML_EXC, msg != null ? msg : "XML data expected");
    } catch (Exception any) {
      throw new SDLIPException(SDLIPException.MALFORMED_XML_EXC, any.getMessage());
    }
  }

  public static XMLObject create(sdlip.XMLObject obj) throws SDLIPException {

    if(obj == null)
      return new XMLObject();
    if(obj instanceof XMLObject)
      return (XMLObject)obj;
    else
      return new XMLObject(obj.getString());
  }

  public static void setElement(sdlip.XMLObject obj, Element el) throws SDLIPException {

    if(obj instanceof XMLObject)
      ((XMLObject)obj).setElement(el);
    else
      obj.setString( DOMWriter.serialize(el) );
  }

  
}
